/*--------------------------------------------
Trace output
  0|A: prints the states of the scanner automaton
  1|F: prints the First and Follow sets of all nonterminals
  2|G: prints the syntax graph of the productions
  3|I: traces the computation of the First sets
  4|J: prints the sets associated with ANYs and synchronisation sets
  6|S: prints the symbol table (terminals, nonterminals, pragmas)
  7|X: prints a cross reference list of all syntax symbols
  8|P: prints statistics about the Coco run
  
Trace output can be switched on by the pragma
     $ {digit|letter}
in the attributed grammar.
--------------------------------------------*/
using System;
using System.IO;

namespace Coco {

public class Coco {
	static StreamWriter trace;

	public static void Main (string[] arg) {
		Console.WriteLine("Coco/R (13-Sep-2002)");
		string ATGName = null; 
		for (int i = 0; i < arg.Length; i++) {
			if (arg[i][0] == '-') Tab.SetDDT(arg[i]); else ATGName = arg[i];
		}
		if (arg.Length > 0 && ATGName != null) {
			Errors.SemErr = new ErrorProc(SemErr);
			Scanner.Init(ATGName);
			OpenTrace();
			Tab.Init(trace); DFA.Init(trace); ParserGen.Init(trace);
			Parser.Parse();
			trace.Close();
			Console.WriteLine("{0} errors detected", Errors.count);
		} else {
			Console.WriteLine("Usage: Coco [-AFGISX] Grammar.ATG [-AFGISX]");
			Console.WriteLine("Options:");
			Console.WriteLine(" -A  trace automaton");
			Console.WriteLine(" -F  list first/follow sets");
			Console.WriteLine(" -G  print syntax graph");
			Console.WriteLine(" -I  trace computation of first sets");
			Console.WriteLine(" -J  print ANY and SYNC sets");
			Console.WriteLine(" -P  print statistics");
			Console.WriteLine(" -S  list symbol table");
			Console.WriteLine(" -X  list cross reference table");
			Console.WriteLine("Scanner.frame and Parser.frame files needed in the current directory");
			Console.WriteLine("or in a directory referenced by the environmen variable CRFRAMES");
		}
	}
	
	static void OpenTrace() {
		FileStream s = new FileStream(System.Environment.CurrentDirectory + "\\" + "listing.txt", FileMode.Create);
		trace = new StreamWriter(s);
	}

	static void SemErr (int n, int line, int col) {
		Console.Write("-- line {0} col {1}: ", line, col);
		Errors.count++;
		string s;
		switch (n) {
			case  1: s = "character set may not be empty"; break;
			case  2: s = "unacceptable character value"; break;
			case  3: s = "a literal must not have attributes"; break;
			case  4: s = "this symbol kind is not allowed in a production"; break;
			case  5: s = "attribute mismatch between declaration and use of this symbol"; break;
			case  6: s = "undefined string in production"; break;
			case  7: s = "name declared twice"; break;
			case  8: s = "this symbol kind not allowed on left side of production"; break;
			case  9: s = "may not ignore \'\\0\'"; break;
			case 11: s = "missing production for grammar name"; break;
			case 12: s = "grammar symbol must not have attributes"; break;
			case 13: s = "a literal must not be declared with a structure"; break;
			case 14: s = "semantic action not allowed here"; break;
			case 15: s = "undefined name"; break;
			case 17: s = "name does not match grammar name"; break;
			case 18: s = "bad string in semantic action"; break;
			case 19: s = "missing end of previous semantic action"; break;
			case 20: s = "token might be empty"; break;
			case 21: s = "token must not start with an iteration"; break;
			case 22: s = "comment delimiters may not be structured"; break;
			case 23: s = "only terminals may be weak"; break;
			case 24: s = "tokens must not contain blanks"; break;
			case 25: s = "comment delimiters must be 1 or 2 characters long"; break;
			case 26: s = "character set contains more than 1 character"; break;
			case 27: s = "\\0 not allowed here. Used as eof character"; break;
			case 28: s = "bad escape sequence in string or character"; break;
			case 29: s = "empty token not allowed"; break;
			default: s = "error " + n; break;
		}
		Console.WriteLine(s);
	}
} // end Coco

} // end namespace